
#include "StdAfx.h"
#include "Plane.h"

#include "Utily.h"
#include "Random.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;

#endif


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MyPlane - 

MyPlane::MyPlane (int nDim) {
	InitMemory(nDim);
	nDimPlane = ::MyRndDimPlane(nDim);
	//    nDimPlane   2, 
	//    -  0  1;
	//   "  " -      2   0/1
	for (int i=0; i<nDim; i++) 
		Array[i] = (::RndLT(double(nDimPlane)/double(nDim))) ? 2 : (RndBool()) ? 0 : 1;
}
	
MyPlane::MyPlane (MyPlane& copy) {
	InitMemory(copy.nDim);
	nDimPlane = copy.nDimPlane;
	for (int i=0; i<nDim; i++)  Array[i] = copy.Array[i];
}

MyPlane::~MyPlane () {
	delete[] Array;
	delete[] CurArray;
}
	
void MyPlane::Init (int nDim, int mPoint) {
	InitMemory(nDim);
	nDimPlane = 1;
	//   Array   0  1
	for (int i=0; i<nDim; i++) {
		Array[i] = mPoint % 2;
		mPoint /= 2;
	}
}

void MyPlane::InitMemory (int nDim) {
	this->nDim = nDim;  BoundsInt(this->nDim,1,maxDim);
	Array = new unsigned char [nDim];
	//         :
	//      
	CurArray = new bool [nDim];
}

bool MyPlane::ExistsPoint (int mPoint) {
	for (int i=0; i<nDim; i++) {
		unsigned char c = mPoint % 2;
		if (!::MyInclude(c,Array[i])) return false;
		mPoint /= 2;
	}
	return true;
}

void MyPlane::SetBoth (int nom) {
	if (Array[nom]==2) ::MyError(20);
	Array[nom] = 2;
	if (++nDimPlane>nDim) ::MyError(2);
}
	
bool operator== (MyPlane& plane1, MyPlane& plane2) {
	int nDim = plane1.nDim;  if (nDim!= plane2.nDim) ::MyError(1);
	for (int i=0; i<nDim; i++)  if (plane1.Array[i]!=plane2.Array[i]) return false;
	return true;
}

bool operator<= (MyPlane& plane1, MyPlane& plane2) {
	int nDim = plane1.nDim;  if (nDim!= plane2.nDim) ::MyError(1);
	for (int i=0; i<nDim; i++)  
		if (!::MyInclude(plane1.Array[i],plane2.Array[i])) return false;
	return true;
}

bool operator< (MyPlane& plane1, MyPlane& plane2) {
	int nDim = plane1.nDim;  if (nDim!= plane2.nDim) ::MyError(1);
	for (int i=0; i<nDim; i++) {
		unsigned char c1 = plane1.Array[i], c2 = plane2.Array[i];
		if (c1==c2) continue;
		return c1<c2;
	}
	return false;
}

istream& operator>> (istream& is, MyPlane& plane) {
	is >> ws >> dec; // -   -   ?
	int nDim;  is >> nDim;
	plane.InitMemory(nDim);
	plane.nDimPlane = 0;
	for (int i=0; i<nDim; i++)   {	// sic!
		unsigned char c;  is >> c;
		if (c<'0' || c>'2') ::MyError(3);
		c -= '0';
		plane.Array[i] = c;
		if (c==2) (plane.nDimPlane)++;
	}
	return is;
}

ostream& operator<< (ostream& os, MyPlane& plane) {
	os << plane.nDim << " ";
	for (int i=0; i<plane.nDim; i++)   os << int(plane.Array[i]);
	return os;
}

void MyPlane::InitSearch () {
	for (int i=0; i<nDim; i++)  CurArray[i] = (Array[i]==1);
	nCurCoord = -1;	// sic!
}
	
int MyPlane::MakeNext () {
	//    ?
	if (nCurCoord>nDim) return -1;
	//  
	int nRet = 0;
	for (int i=nDim-1; i>=0; i--) {
		nRet *= 2;
		if (CurArray[i]) nRet += 1;
	}
	//     
	for (;;) {
		//          0  1?
		if (nCurCoord>=0 && !CurArray[nCurCoord] && Array[nCurCoord]==2) {
			CurArray[nCurCoord] = true;	// 
			//     :
			for (int i=nCurCoord-1; i>=0; i--) if (Array[i]==2) {
				CurArray[i] = false;
				nCurCoord = i;
			}
			break;
		}
		if (++nCurCoord>nDim) break;	
		//  (    )    () 
	}
	// 
	return nRet;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MyArrayPlane - 

MyArrayPlane::~MyArrayPlane () {
    for (int i=0; i<GetSize(); i++) {
        MyPlane* pPlane = (MyPlane*)GetAt(i);
        delete pPlane;
    }
}
	
int MyArrayPlane::ExistsPlane (MyPlane* plane) {
    for (int i=0; i<GetSize(); i++) {
        MyPlane* pPlane = (MyPlane*)GetAt(i);
		if (*plane == *pPlane) return i;
		if (*plane <= *pPlane) return -1;
	}
	return -2;
}
	
int MyArrayPlane::PlaneByPoint (int mPoint) {
    for (int i=0; i<GetSize(); i++) {
        MyPlane* pPlane = (MyPlane*)GetAt(i);
		if (pPlane->ExistsPoint(mPoint)) return i;
	}
	return -2;
}

ostream& operator<< (ostream& os, MyArrayPlane& arrplane) {
    for (int i=0; i<arrplane.GetSize(); i++) {
        MyPlane* pPlane = (MyPlane*)arrplane.GetAt(i);
        os << *pPlane << endl;
    }
	return os;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MyArrayPlane16 - 

MyArrayPlane16::MyArrayPlane16 (int mDim) {
	this->mDim = mDim;
	this->kolPlanes = 0;
}

void MyArrayPlane16::AddPlane (MyPlane* plane) {
	if (kolPlanes>=maxPlanes) ::MyError(19);
	for (int m=0; m<mDim; m++) Array[kolPlanes][m] = false;
	plane->InitSearch();
	for (;;) {
		int m = plane->MakeNext();
		if (m<0) break;
		Array[kolPlanes][m] = true;
	}
	kolPlanes++;
}
