
#include "StdAfx.h"
#include "Qube.h"
#include "Utily.h"
#include "Random.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;

#endif


const int Degree[32] =
	{ 1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024,  //    0  10
	  2048, 4096, 8192, 16384, 32768, 65536,		//   11  16
	  131072, 262144, 524288, 1048576, 2097152,     //   17  21
	  4194304, 8388608, 16777216, 33554432,			//   22  25
	  67108864, 134217728, 268435456,				//   26  28
	  536870912, 1073741824, 2147483648			    //   29  31
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MyQube - 
	
MyQube::MyQube (int nDim) {
	InitMemory(nDim);
}
		
MyQube::MyQube (int nDim, int kolPlane) {
	InitMemory(nDim);
	for (int i=0; i<kolPlane; i++) AddRandomPlane();
	MakeCost();
}
	
MyQube::MyQube (MyQube& copy) {
	this->nDim = copy.nDim;
	InitMemory(this->nDim);
	this->mDim = copy.mDim;
	for (int m=0; m<mDim; m++)  this->Massiv[m] = copy.Massiv[m];
	this->nSumma = copy.nSumma;
	for (int i=0; i<nDim; i++)  this->Cost[i] = copy.Cost[i];
}

MyQube::~MyQube () {
	//delete[] Array;
	delete[] Massiv;
	delete[] Cost;
}

void MyQube::InitMemory (int nDim) {
	this->nDim = nDim;  BoundsInt(this->nDim,1,maxDim);
	mDim = Degree[nDim];
	//nDimension = Degree[nDim-3];
	//Array = new unsigned char [nDimension];
	Massiv = new bool [mDim];
	//for (int i=0; i<nDimension; i++)  Array[i] = 0;
	for (int m=0; m<mDim; m++)  Massiv[m] = false;
	nSumma = 0;
	Cost = new int[nDim];
}

void MyQube::SetTrue (int mPoint) {
	if (mPoint<0 || mPoint>mDim) ::MyError(4);
	if (Get(mPoint)) return;
	//Array[mPoint/8] |= Degree[mPoint%8];
	Massiv[mPoint] = true;
	nSumma++;
}

bool MyQube::Get (int mPoint) {
	if (mPoint<0 || mPoint>mDim) ::MyError(4);
	//return (Array[mPoint/8] & Degree[mPoint%8]) != 0;
	return Massiv[mPoint];
}
	
void MyQube::AddPlane (MyPlane& plane) {
	plane.InitSearch();
	for (;;) {
		int mPoint = plane.MakeNext();
		if (mPoint<0) return;
		SetTrue(mPoint);
	}
}

void MyQube::AddRandomPlane () {
	MyPlane planeRnd(nDim);
	AddPlane(planeRnd);
}
	
void MyQube::MakeCost () {
	int i;
	for (int i=0; i<nDim; i++) Cost[i] = 0;
	for (int m=0; m<GetMDim(); m++) {       //   
		if (!Get(m)) continue;
		for (i=0; i<nDim; i++) {		//      
			int mmm = ::Sosed(m,nDim,i);
			if (Get(mmm)) Cost[i]++;
		}
	}
	for (i=0; i<nDim; i++) {
		if (Cost[i]%2==1) ::MyError(7);
		Cost[i] /= 2;
	}
}

int MyQube::RandomPoint () {
	RndInt arrRnd;
	for (int m=0; m<mDim; m++)  if (Get(m))  arrRnd.Add(m);
	return arrRnd.GetAnswer();
}

istream& operator>> (istream& is, MyQube& qube) {
	is >> ws >> dec; // -   -   ?
	int nDim;  is >> nDim;
	qube.InitMemory(nDim);
	for (int m=0; m<qube.mDim; m++) {       // sic!
		unsigned char c;  is >> c;
		if (c=='1') qube.SetTrue(m);
		else if (c!='0') ::MyError(3);
	}
	qube.MakeCost();
	return is;
}

ostream& operator<< (ostream& os, MyQube& qube) {
	os << qube.nDim << " ";		 // sic!
	for (int m=0; m<qube.mDim; m++) // sic!
		if (qube.Get(m)) os << 1;
		else  os << 0;
	return os;
}
	
void MyQube::ReadFile (CString& strName) {
	ifstream FileIn;
	FileIn.open(strName,ios::_Nocreate);
	if (!FileIn) ::MyError(5);
	FileIn >> *this;
	FileIn.close();
}
	
void MyQube::WriteFile (CString& strName) {
	ofstream FileOut;
	FileOut.open(strName);
	if (!FileOut) ::MyError(5);
	FileOut << *this;
	FileOut.close();
}
	
int MyQube::HowManyNew (MyPlane& plane) {
	int nRet = 0;
	for (int m=0; m<mDim; m++) {    //   
		if (!Get(m)) continue;
		if (!plane.ExistsPoint(m)) nRet++;
	}
	return nRet;
}
	
bool MyQube::CanExtand (MyPlane& plane) {
	for (int nCoord=0; nCoord<nDim; nCoord++) {
		if (plane[nCoord]==2) continue;
		//  !   MyPlane    []
		//       
		plane.InitSearch();
		for (;;) {
			int m = plane.MakeNext();
			if (m<0) return true;	//        
			int mmm = ::Sosed(m,nDim,nCoord);
			if (!Get(mmm)) break;   //     -   
		}
	}
	return false;
}

int MyQube::ExtandRandom (MyPlane& plane) {
	int nAdd = ::Rnd(nDim);
	for (int i=0; i<nDim; i++) {
		int nCoord = (i+nAdd) % nDim;   //    
		if (plane[nCoord]==2) continue;
		plane.InitSearch();
		for (;;) {
			int m = plane.MakeNext();
			if (m<0) return nCoord; //        
			int mmm = ::Sosed(m,nDim,nCoord);
			if (!Get(mmm)) break;   //     -   
		}
	}
	return -1;
}

/*
int MyQube::ExtandGreedy (MyPlane& plane) {
	if (::RndLT(0.1)) {
		int n = ExtandGreedy3(plane);
		if (n>=0) return n;
	}
	int n = ExtandGreedy2(plane);
	if (n>=0) return n;
	return ExtandRandom(plane);
}
	
int MyQube::ExtandGreedy2 (MyPlane& plane) {
	MySuperSet pairs;
	for (int n1=0; n1<nDim-1; n1++)  {
		if (plane.Get(n1)==2) continue;
		//          n1
		for (int n2=n1+1; n2<nDim; n2++)  {
			if (plane.Get(n2)==2) continue;
			//   ,       , n1  n2
			//       (,  1)  Ш 
			plane.InitSearch();
			for (;;) {
				int m = plane.MakeNext();
				if (m<0) {				//        
					pairs.MyAdd(n1);
					pairs.MyAdd(n2);
					break;				//     
				}
				int m1 = ::Sosed(m,nDim,n1);
				if (!Get(m1)) break;	//     
				int m2 = ::Sosed(m,nDim,n2);
				if (!Get(m2)) break;	//     
				int mmm = ::Sosed(m1,nDim,n2);
				if (mmm!=::Sosed(m2,nDim,n1)) ::MyError(21);
				if (!Get(mmm)) break;	//     
			}
		}
	}
	return pairs.MyGet();
}
	
int MyQube::ExtandGreedy3 (MyPlane& plane) {	//  
	MySuperSet pairs;
	for (int n1=0; n1<nDim-2; n1++)  {
		if (plane.Get(n1)==2) continue;
		for (int n2=n1+1; n2<nDim-1; n2++)  {
			if (plane.Get(n2)==2) continue;
			for (int n3=n2+1; n3<nDim; n3++)  {
				if (plane.Get(n3)==2) continue;
				plane.InitSearch();
				for (;;) {
					int m = plane.MakeNext();
					if (m<0) {				//        
						pairs.MyAdd(n1);
						pairs.MyAdd(n2);
						pairs.MyAdd(n3);
						break;				//     
					}
					int m1 = ::Sosed(m,nDim,n1);
					if (!Get(m1)) break;	//     
					int m2 = ::Sosed(m,nDim,n2);
					if (!Get(m2)) break;	//     
					int m3 = ::Sosed(m,nDim,n3);
					if (!Get(m3)) break;	//     
					int m12 = ::Sosed(m1,nDim,n2);
					if (!Get(m12)) break;	//     
					int m13 = ::Sosed(m1,nDim,n3);
					if (!Get(m13)) break;	//     
					int m23 = ::Sosed(m2,nDim,n3);
					if (!Get(m23)) break;	//     
					int m123 = ::Sosed(m12,nDim,n3);
					if (!Get(m123)) break;	//     
				}
			}
		}
	}
	return pairs.MyGet();
}
*/
	
MyPlane* MyQube::MakePlaneRnd (int mPoint) {
	if (!this->Get(mPoint)) return NULL;
	MyPlane* planeNew = new MyPlane(this->GetDim(),mPoint);
	for (;;) {
		int nom = this->ExtandRandom(*planeNew);
		if (nom<0) return planeNew;
		planeNew->SetBoth(nom);
	}
}
	
/*
MyPlane* MyQube::MakePlaneGreedy (int mPoint) {
	if (!this->Get(mPoint)) return NULL;
	MyPlane* planeNew = new MyPlane(this->GetDim(),mPoint);
	for (;;) {
		int nom = this->ExtandGreedy(*planeNew);
		if (nom<0) return planeNew;
		planeNew->SetBoth(nom);
	}
}

int operator- (MyQube& qube1, MyQube& qube2) {
	int mDim = qube1.GetMDim();  if (mDim!=qube2.GetMDim()) ::MyError(8);
	int nRet = 0;
	for (int m=0; m<mDim; m++) {    //   
		if (!qube1.Get(m)) continue;
		if (!qube2.Get(m)) nRet++;
	}
	return nRet;
}
*/

void operator>> (MyQube& qube1, MyQube& qube2) {
	int i;
	//         
	// (    ) -
	// ,     (   )?
	int nDim = qube1.GetDim();
	qube2.InitMemory(nDim);		//     false
	int* NomCoordNew = new int[nDim];	//   
	for (int i=0; i<nDim; i++)  NomCoordNew[i] = i;
	for (i=0; i<1000; i++)  {
		int n1 = ::Rnd(nDim);
		int n2 = ::Rnd(nDim);
		::Obmen(NomCoordNew[n1],NomCoordNew[n2]);
	}
	bool* CoordOld = new bool[nDim];	//   
	int mDim = qube1.GetMDim();
	for (int m=0; m<mDim; m++) {		// m -  
		if (!qube1.Get(m)) continue;
		int mmm = m;	//    ,   m
		for (int i=0; i<nDim; i++) {
			CoordOld[i] = (mmm % 2 == 1);
			mmm /= 2;
		}
		mmm = 0;		//   -   "m-"   
		int nnn = 1;
		for (i=0; i<nDim; i++) {
			if (CoordOld[NomCoordNew[i]]) mmm += nnn;
			nnn *= 2;
		}
		qube2.SetTrue(mmm);
	}
	delete NomCoordNew;
	delete CoordOld;
	if (qube1.GetSumma()!=qube2.GetSumma()) ::MyError(13);
}
